package org.eclipse.palamedes.gdl.connection;

import java.util.ArrayList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;


/**
 * This class handles all message related transformations.
 * @author ingo
 *
 */

// TODO: use KIF parser instead of this handwritten stuff here

public class Message {

    /* Message strings */
    static final String STR_START = "start";
    static final String STR_STOP  = "stop";
    static final String STR_PLAY  = "play";

    /* Message types */
    static final int ERROR = -1;
    static final int EMPTY = 0;
    static final int START = 1;
    static final int PLAY  = 2;
    static final int STOP  = 3;

    /* Regular Expressions */
    static final Pattern RE_COMMAND = Pattern.compile("\\(([^\\s]*)\\s");

    /* Message content fields */
    int      type;
    String   matchId;
    String   role;
    String   gameDescription;
    String[] moves;
    int      startClock;
    int      playClock;
    String   originalMessage;


    /**
    * Constructs a message object from a message string.
    * @param message
    */
    public Message(String message) {        
        originalMessage = message;
        parseMessage(originalMessage);
    }


    /**
    * This method parses the message string into a message object.
    * @param message
    */
    public void parseMessage(String message) {

        // resets the message
        reset();

        // no message there
        if ( message == null )
            return;

        message=message.toLowerCase().trim();
        
        // retrieve the command
        String   command      = getCommand( message );
        if (command == null)
            throw new RuntimeException("No command found in message: "+message);
            
        // remove closing brackets
        message = message.substring( 1, message.length() - 1 );

        // split message in its parts
        String[] messageParts = message.split( " " );
        int      length       = messageParts.length;

        // create temp string for move parsing
        String   tmpString    = "";

        /* START Message handling */
        if ( command.equals( Message.STR_START ) ) {
            type       = Message.START;
            matchId    = messageParts[1];
            role       = messageParts[2];
            startClock = Integer.parseInt( messageParts[length-2] );
            playClock  = Integer.parseInt( messageParts[length-1] );

            // all other groups are related to the game description
            for (int i = 3; i < length - 2; i++)
                tmpString += " " + messageParts[i];

            tmpString       = tmpString.trim();
            tmpString       = tmpString.substring(1, tmpString.length() - 1);
            gameDescription = tmpString;

        /* PLAY Message handling */
        } else if ( command.equals( Message.STR_PLAY ) ) {

            type    = Message.PLAY;
            matchId = messageParts[1];

            // all other groups are related to the move section
            for (int i = 2; i < messageParts.length; i++)
                tmpString += " " + messageParts[i];
            System.out.println("parsing moves from \""+tmpString+"\"");
            moves     = Message.getMoveSubstrings(tmpString.trim());

        /* STOP Message handling */
        } else if ( command.equals( Message.STR_STOP ) ) {

            type    = Message.STOP;
            matchId = messageParts[1];

            // STOP Message does not have to have a move list
            try {
                // all other groups are related to the move section
                for (int i = 2; i < messageParts.length; i++)
                    tmpString += " " + messageParts[i];

                moves     = Message.getMoveSubstrings(tmpString.trim());

            } catch ( Exception ex ) {
                moves = null;
            }

        } else {
            reset();
            type = Message.ERROR;
            //logger.error( "Command not found: " + command );
        }

        //logger.debug( "Finished parsing the message." );
    }


    /**
     * This method extracts the message command string.
     */
    private String getCommand(String message) {
        String command = null;
        try {

            Matcher m = RE_COMMAND.matcher(message);
            if ( m.lookingAt() ) {
                command = m.group(1);
            }

        } catch (Exception ex) {
            //logger.error( "Pattern to extract command did not match!", ex );
            this.type = Message.ERROR;
        }

        return command;
    }

    /**
     * This method parses a GDL string into a ExpressionList.
     * If a error occurs while parsing the result will be an empty list.
     * @param gdl
     * @return
     */
    public static String[] getMoveSubstrings( String gdl ) {

        // the first answer in a match
        if ( gdl.equals( "nil" ) )
            return new String[0];

        // remove additional outer brackets -> used for move parsing
        gdl = gdl.substring(1, gdl.length() - 1 ).trim();
        // parse stream into game description
        return splitMovesString(gdl);
    }


    /**
     * Separates the moves through bracket counting. If bracket count hits 0 a
     * new move string is assumed to be started.
     */
    static String[] splitMovesString(String moves) {
        int           count    = 0;
        int           lastCut  = 0;
        List<String>  movesStr = new ArrayList<String>();
        int i, j;
        
        for (i = 0, j = moves.length(); i < j; i++){
            switch (moves.charAt(i)) {
            case '(':
                count += 1;
                break;
            case ')':
                count -= 1;
                break;
            case ' ':
            	// do the cut
                if (count == 0) {
                    String move = moves.subSequence(lastCut, i).toString();
                    movesStr.add( move.trim() );
                    lastCut = i+1;
                }
            }
        }
        if(lastCut<i){
            String move = moves.subSequence(lastCut, i).toString();
            movesStr.add( move.trim() );
        }
        	
        return movesStr.toArray( new String[movesStr.size()] );
    }

    /**
    * This method handles a clean reset of the message object. Essentially it
    * sets all content fields to a default empty state.
    */
    private void reset() {
        this.type    		 = Message.EMPTY;
        this.matchId 		 = "";
        this.role    		 = "";
        this.gameDescription = "";
        this.moves   		 = null;
        this.startClock		 = 0;
        this.playClock	     = 0;
    }

    @Override
    public String toString() {
        return originalMessage;
    }    
    

    /**
     * @return the gameDescription
     */
    public String getGameDescription() {
        return gameDescription;
    }


    /**
     * @param gameDescription the gameDescription to set
     */
    public void setGameDescription(String gameDescription) {
        this.gameDescription = gameDescription;
    }


    /**
     * @return the matchId
     */
    public String getMatchId() {
        return matchId;
    }


    /**
     * @param matchId the matchId to set
     */
    public void setMatchId(String matchId) {
        this.matchId = matchId;
    }


    /**
     * @return the moves
     */
    public String[] getMoves() {
        return moves.clone();
    }

    /**
     * @return whether the message contains moves or not
     */
    public boolean hasMoves() {
        return moves != null;
    }

    /**
     * @param moves the moves to set
     */
    public void setMoves(String[] moves) {
        this.moves = moves.clone();
    }


    /**
     * @return the playClock
     */
    public int getPlayClock() {
        return playClock;
    }


    /**
     * @param playClock the playClock to set
     */
    public void setPlayClock(int playClock) {
        this.playClock = playClock;
    }


    /**
     * @return the role
     */
    public String getRole() {
        return role;
    }


    /**
     * @param role the role to set
     */
    public void setRole(String role) {
        this.role = role;
    }


    /**
     * @return the startClock
     */
    public int getStartClock() {
        return startClock;
    }


    /**
     * @param startClock the startClock to set
     */
    public void setStartClock(int startClock) {
        this.startClock = startClock;
    }


    /**
     * @return the type
     */
    public int getType() {
        return type;
    }


    /**
     * @param type the type to set
     */
    public void setType(int type) {
        this.type = type;
    }

}
