package org.eclipse.palamedes.gdl.connection;

import java.util.Arrays;

import org.eclipse.palamedes.gdl.core.model.GameFactory;
import org.eclipse.palamedes.gdl.core.model.IGame;
import org.eclipse.palamedes.gdl.core.model.utils.Game;
import org.eclipse.palamedes.gdl.core.simulation.IStrategy;
import org.eclipse.palamedes.gdl.core.simulation.Match;
import org.eclipse.palamedes.gdl.core.simulation.strategies.SMonteCarloUCT;

public class Player {

    protected final static String STR_ERROR1 = "Play message for the wrong match ";

    protected static class RealMatch extends Match {

        final String matchId;

        public RealMatch( String    matchId,
                          IGame     game,
                          IStrategy strategy,
                          String    role,
                          int       start,
                          int       play ) {
            super(game, strategy, role, start, play);
            this.matchId = matchId;
        }

    }

    /* GameSimulator of the player */
    protected RealMatch realMatch;


    /**
     * This method is called when a new match begins.<br>
     * <br>
     * msg="(START &lt;MATCH ID&gt; &lt;ROLE&gt; &lt;GAME DESCRIPTION&gt; &lt;STARTCLOCK&gt;
     *      &lt;PLAYCLOCK&gt;)"<br>
     * e.g. msg="(START tictactoe1 white ((role white) (role black) ...) 1800
     *          120)" means:
     * <ul>
     *   <li>the current match is called "tictactoe1"</li>
     *   <li>your role is "white",</li>
     *   <li>
     *     after at most 1800 seconds, you have to return from the
     *     commandStart method
     *   </li>
     *   <li>for each move you have 120 seconds</li>
     * </ul>
     *
     * TODO:
     *    - use the time to "contemplate" about the game description
     *      and return on time (before STARTCLOCK is over!)
     */
    public void commandStart(Message msg){

        // create the clock
        int playClock  = msg.getPlayClock()  - 1;
        int startClock = msg.getStartClock() - 1;

        System.out.println( "Start Clock " + startClock );
        System.out.println( "Play  Clock " + playClock  );

        // get the game from the database
        GameFactory factory = GameFactory.getInstance();
        IGame runningGame   = factory.createGame( GameFactory.PROLOGPROVER,
                                                    msg.getGameDescription() );
        System.out.println("Bender created the game.");

        IStrategy strategy = new SMonteCarloUCT();
        System.out.println( "Bender created the strategy "      +
                            strategy.getClass().getSimpleName() +
                            "." );

        System.out.println( "Bender starts contemplate while doing yoga." );

        // create a match
        realMatch = new RealMatch( msg.getMatchId(),
                                   runningGame,
                                   strategy,
                                   msg.getRole(),
                                   startClock,
                                   playClock );
        System.out.println( "Bender created the match." );
        System.out.println( "Bender is prepared to start the game." );
        System.out.println("stats:"+runningGame.getStatistic());
    }

    /**
     * This method is called once for each move<br>
     * <br>
     * msg = "(PLAY <MATCH ID> <JOINT MOVE>)<br>
     * <JOINT MOVE> will be NIL for the first PLAY message and the list of the
     * moves of all the players in the previous state<br>
     * e.g. msg="(PLAY tictactoe1 NIL)" for the first PLAY message
     *   or msg="(PLAY tictactoe1 ((MARK 1 2) NOOP))" if white marked cell (1,2)
     *   and black did a "noop".<br>
     * <br>
     * TODO:<br>
     * <ul>
     *   <li>
     *     calculate the new state from the old one and the <JOINT MOVE>
     *   </li>
     *   <li>
     *     use the time to find the best of your possible moves in the current
     *     state
     *   </li>
     *   <li>
     *     return your move (instead of "NIL") on time (before PLAYCLOCK is
     *     over!)
     *   </li>
     * </ul>
     * @return the move of this player
     */
    public String commandPlay(Message msg){
        checkMatchId(msg);

        // only make a turn if the move list is not empty
        // is it empty it means we hit the first play message, setting the
        // initial state is done while constructing the match object
        if ( msg.hasMoves() ){
            System.out.println( "Moves from GameMaster: " +
                                Arrays.toString(msg.getMoves()) );

            String[] prepared = prepareMoves(msg);

            // got the initial NIL in case of length == 0
            if (prepared.length != 0)

                // prepare moves
                realMatch.makeTurn( prepared );

        }

        // real work is done here
        String move = realMatch.getMove().getMove().toUpperCase();

        // logging the resulting move
        System.out.println( "Our move: " + move );
        System.out.println("stats:"+((Game)realMatch.getGame()).getStatistic());
        return move;
    }


    /**
     * This method is called if the match is over
     *
     * msg="(STOP <MATCH ID> <JOINT MOVE>)
     *
     * TODO:
     * <ul>
     *   <li>clean up the GamePlayer for the next match</li>
     *   <li>
     *     be happy if you have won, think about what went wrong if you have
     *     lost ;-)
     *   </li>
     * </ul>
     */
    public void commandStop(Message msg){
        checkMatchId(msg);

        // adds the moves to the match
        if ( msg.hasMoves() )

            // real work is done here
            realMatch.makeTurn( prepareMoves(msg) );

        // check if we agree to be in a final state and log the result
        if ( !realMatch.getCurrentNode().isTerminal() )
            System.out.println( "Game stopped but not finished" );
        else{
            System.out.println( "Game finished" );
			try {
	        	int[] goalvalues;
				goalvalues = realMatch.getGame().getState(realMatch.getCurrentNode()).getGoalValues();
	        	System.out.println( "goal:" );
	        	for(int i=0;i<goalvalues.length;++i){
	        		System.out.print( goalvalues[i]+" " );
	        	}
	        	System.out.println();
			} catch (InterruptedException e) {
				System.out.println("Timeout while computing goal values:");
				e.printStackTrace();
			}
        }
    }


    /**
     * This method is for error checking nothing more.
     * @param msg
     */
    protected void checkMatchId( Message msg ) {
        if (realMatch == null)
            throw new RuntimeException("Found no running match.");
        
        else if (msg == null)
            throw new RuntimeException("Null message received.");
        
        if ( !realMatch.matchId.equals( msg.getMatchId() ) )
            System.out.println( STR_ERROR1 + msg.getMatchId() );
    }


    protected String[] prepareMoves(Message msg) {
        String[] moves = msg.getMoves();
        String[] names = realMatch.getGame().getRoleNames();
        for (int i = 0, j = moves.length; i < j; i++) {
            String tmp = moves[i].trim();
            // add player name
            moves[i] = "(does " + names[i] + " " + tmp + ")";
        }
        return moves;
    }

    public double getReliabilityValue() {
        return (realMatch != null) ? realMatch.getReliabilityValue() : 0.0;
    }
    
    public double getExpectedGoalValue() {
        return realMatch != null ? 
                                realMatch.getStrategy().getExpectedGoalValue() : 
                                0.0;
    }

    public String getSlaveExtention() {
        return "#" + getExpectedGoalValue() + "#" + getReliabilityValue();
    }

    public RealMatch createRealMatch( String    matchId,
            						  IGame     game,
            						  IStrategy strategy,
            						  String    role,
            						  int       start,
            						  int       play ) {
    	return new RealMatch(matchId, game, strategy, role, start, play);
    }
    
}
