package org.eclipse.palamedes.gdl.connection;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.net.Socket;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.List;
import java.util.Properties;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

public class PlayerProxy extends NanoHTTPD {
    
    
    private static final class MessageForwarder implements Callable<String> {
     
        private final String msg;
        private final String host;
        private final int    port;
        
        MessageForwarder(String msg, String host, int port) {
            this.msg  = msg;
            this.host = host;
            this.port = port;
            System.out.println("Created forwarder to " + host + ":" + port);
        }
        
        private String sendMsg(String msgStr, String hostStr, int portNum) {        
            String reply = null;
            
            try {
                Socket       socket = new Socket(hostStr, portNum);            
                OutputStream out    = socket.getOutputStream();
                PrintWriter  pw     = new PrintWriter(out);
                pw.print("POST / HTTP/1.0\r\n");
                pw.print("Accept: text/delim\r\n");
                pw.print("Sender: ProxyPlayer\r\n");
                pw.print("Receiver: " + hostStr + "\r\n");
                pw.print("Content-type: text/acl\r\n"); 
                pw.print("Content-length: "+ msgStr.length() + "\r\n");   
                pw.print("\r\n");   
                pw.print(msgStr);
                pw.flush();

                InputStream is = socket.getInputStream();
                if (is == null) 
                    return null;
      
                BufferedReader in = new BufferedReader( new InputStreamReader(is) );
                String line;
                line = in.readLine();
                while( line != null && 
                       line.trim().length() > 0 )
                    line = in.readLine();
        
                char[] cbuf = new char[1024];
                int len;
                while( (len = in.read(cbuf,0,1023)) != -1 )
                    reply = (reply == null) ? 
                                              new String(cbuf, 0, len) : 
                                              reply + new String(cbuf, 0, len); 

                out.close();
                in.close();
            } catch (UnknownHostException e) {
                System.err.println( "error: unknown host [" + hostStr + "]" );
            } catch (IOException e) {
                System.err.println( "error: io error for " + 
                                    this + " : " + e.getMessage() );
            }
            return reply;
        }
        
        public String call() throws Exception {
            System.out.println("send to [" + host + ":" + port + "] -> " + msg);             
            return sendMsg(msg.toString(), host, port);
        }

    }
    
    /* Internal used strings */
    final static String STR_SERVER_START = "Started a PlayerServer: ";
    final static String STR_ERROR1       = "Message is empty!";
    final static String STR_ERROR2       = "Unknown command ";
    final static String STR_ERROR3       = "Unknown message state ";

    final int[]    playerPorts; 
    final String[] playerHosts;
    
    /** Offset for the timeout calculation in milliseconds. */
    private static final int timeoutOffset = 0;

    /** Timeout for the start message in milliseconds. */
    private int startTime;
    
    /** Timeout for the play and stop messages in milliseconds. */
    private int playTime;
    
    /** Responses collect all the responses from ply*/
    List<Future<String>> responses = new ArrayList<Future<String>>();

    /** Thread Pool */
    ExecutorService threadPool = Executors.newCachedThreadPool();
    
    public PlayerProxy(int port, String[] hosts, int[] ports) 
    throws IOException {
        super(port);
        playerHosts = hosts;
        playerPorts = ports;        
    }

    
    /**
     * Starts the game players and waits for messages from the game master <br>
     * Command line options: [port]
     */
    public static void main(String[] args){
        try {
            int      port        = Integer.parseInt(args[0]);            
            int[]    playerPorts = new int[args.length-1];
            String[] hosts       = new String[args.length-1];            
            for (int i = 0, j = args.length -1; i < j; i++) {
                String[] split = args[i+1].split(":");
                hosts[i]       = split[0];
                playerPorts[i] = Integer.parseInt(split[1]);
            }
            
            // create the proxy and run it
            new PlayerProxy(port, hosts, playerPorts).waitForExit();
            
        } catch (IOException ex) {
            ex.printStackTrace();
            System.exit(-1);
        
        }  catch (Exception ex) {
            System.out.println("Wrong program arguments!");
            printUsage();
            System.exit(-1);
        }
        
    }

    public static final void printUsage() {
        String usage = "\n" +
        "Usage: PlayerProxy [PROXY_PORT] { [HOST]:[PORT] }*\n" +
        "Choose between different PlayerServer at runtime\n" +
        "Example: PlayerProxy 5000 localhost:4001 192.168.0.1:10000";
        System.out.println(usage);
    }
    
    
    /**
     * Wait for server thread to exit
     */
    public void waitForExit(){
        try {
            server_thread.join();
        } catch (Exception ex) {
            ex.printStackTrace();
        }
    }

    
    @Override
    public Response serve( String uri, String method, Properties header, Properties parms, String data ) {
        try {
            long   start           = System.currentTimeMillis();
            String response_string = "";

            if ( data != null ) {

                // send message to player
                responses.clear();
                for (int i = 0; i < playerHosts.length; i++) 
                    responses.add( 
                            threadPool.submit(
                                    new MessageForwarder( data, 
                                                          playerHosts[i], 
                                                          playerPorts[i] ) ) );

                // parse message
                Message message = new Message(data);
                
                // switch to correct function
                switch ( message.type ) {

                case Message.START:
                    startTime    = message.startClock * 1000 - timeoutOffset;
                    playTime     = message.playClock  * 1000 - timeoutOffset;
                    response_string = waitForMessages(start, startTime).get(0);
                    break;

                case Message.PLAY:
                    response_string = chooseBestAnswer(
                                             waitForMessages(start, playTime) );
                    break;
                
                case Message.STOP:                    
                    response_string = waitForMessages(start, playTime).get(0);
                    break;

                case Message.EMPTY:
                    throw( new IllegalArgumentException(STR_ERROR1) );
                
                case Message.ERROR:
                    throw( new IllegalArgumentException(STR_ERROR2 + data) );
                
                default:
                    throw( new IllegalArgumentException(STR_ERROR3 + data) );
                }
                
            } else 
                System.out.println( "Empty message received!" );

            // debug output
            System.out.println( "Response:" + response_string );

            Response response = new Response( Response.HTTP_OK, "text/acl", 
                                              response_string );
            response.addHeader("Content-Length", "" + response_string.length());
            return response;

        // error handling
        } catch ( IllegalArgumentException e ) {
            e.printStackTrace();
        }
        catch (InterruptedException e) {
            e.printStackTrace();
        }
        catch (ExecutionException e) {
            e.printStackTrace();
        }
        catch (TimeoutException e) {
            e.printStackTrace();
        }
        return new Response( Response.HTTP_BADREQUEST, "text/acl", "NIL" );
    }


    /**
     * Wait for incoming responses to sent messages.
     */
    private final List<String> waitForMessages(long start, long timeout) 
    throws InterruptedException, ExecutionException, TimeoutException {
        
        List<String> realResponses = new ArrayList<String>();
        for (Future<String> futureObj : responses) {
            
            long timeoutLeft = timeout - (System.currentTimeMillis() - start);
            if (timeoutLeft > 0)
            
                // wait for the next response
                realResponses.add( futureObj.get( timeoutLeft, 
                                                  TimeUnit.MILLISECONDS ) );            
        }        
        return realResponses;
    }


    /**
     * Chooses the best answer based on the reliability value.
     */
    private final String chooseBestAnswer(List<String> strings) {
        System.out.println("Moves to choose from:");
        double bestValue    = -1;
        double bestVariance = -1;
        String text         = null;
        for (String message : strings) {
            System.out.println(message);
            String[] parts    = message.split("#");
            double   value    = Double.parseDouble(parts[1]);
            double   variance = Double.parseDouble(parts[1]);
            if ( value > bestValue || 
                (value == bestValue && variance < bestVariance) ) {
                bestValue    = value;
                bestVariance = variance;
                text         = parts[0];
            }
        }

        // maybe we did not found a best value
        return (text != null) ? text : strings.get(0).split("#")[0];
    }

}
