package org.eclipse.palamedes.gdl.core.model;

import java.util.List;

/**
 * The IGame interface follows the facade design pattern. It hides the game 
 * information storage, the game analysis, the reasoning machinery and 
 * the game state space representation.
 */
public interface IGame {

    /**
     * Returns the original GDL description for this game.
     */
    public String getSourceGDL();

    /**
     * Returns the string representation of the role names as string array.
     */
    public String[] getRoleNames();

    /**
     * Returns the number of roles respective the number of players.
     */
    public int getRoleCount();

    /**
     * Returns the index of the given role in the role name array.
     */
    public int getRoleIndex(String role);

    /**
     * Returns the game tree representation if it exists; otherwise null.
     */
    public IGameTree getTree();
    
    /**
     * Returns the reasoner that is used for this this game.
     */
    public IReasoner getReasoner();
    
    /**
     * Returns a list of all moves combined.
     * @throws InterruptedException
     */
    public List<IMove[]> getCombinedMoves(IGameNode currentNode)
    throws InterruptedException;

    /**
     * Returns an array of the legal moves of each player. <br/>
     * e.g.: getLegalMoves(currentNode)[playerNumber] is an array of the
     * legal moves of the specified player in the current state 
     * @throws InterruptedException
     */
    public IMove[][] getLegalMoves(IGameNode currentNode)
    throws InterruptedException;

    /**
     * Returns a random combined move.
     * @throws InterruptedException
     */
    public IMove[] getRandomMove(IGameNode currentNode)
    throws InterruptedException;

    /**
     * Returns the child node of the parent that is connected through the
     * specified moves.
     * @throws InterruptedException
     */
    public IGameNode getNextNode(IGameNode currentNode, IMove[] moves)
    throws InterruptedException;

    /**
     * Returns true if the state of the node is a terminal state 
     * @param currentNode
     * @throws InterruptedException
     */
    public boolean isTerminal(IGameNode currentNode)
    throws InterruptedException;

    /**
     * Returns an array of the goal values of all players.
     * A value of -1 is returned if there is no goal value for a player in this state.
     * This may happen if the state is not terminal. 
     * @throws InterruptedException
     */
    public int[] getGoalValues(IGameNode currentNode)
    throws InterruptedException;

    /**
     * Returns the game state associated with currentNode. Regenerates the node if necessary.
     * @see IGame#regenerateNode(IGameNode node)
     * @throws InterruptedException
     */
    public IGameState getState(IGameNode currentNode)
    throws InterruptedException;

    /**
     * This method regenerates the game states if the game node lost its state
     * information for one reason or another.
     * @throws InterruptedException
     */
    public void regenerateNode(IGameNode node)
    throws InterruptedException;

    /**
     * Returns the maximum number of steps if this information is known from the
     * rules; otherwise -1;
     */
    public int getMaxSteps();
    
    public IStatistic getStatistic();

    /**
     * Returns the perft(n) value starting from the given startNode. 
     * 
     * The perft value is defined as the number of game nodes reachable in 
     * exactly n steps.
     */
	public int getPerftValue(int n, IGameNode startNode)
    throws InterruptedException;
	

	/**
	 * Initialises a game object. This method gets called by the game factory
	 * while creating the game.
	 * @param gdl
	 */
	public void init(String gdl);

}
