package org.eclipse.palamedes.gdl.core.model.utils;

import java.util.ArrayList;
import java.util.List;

/**
 * The AbstractFactory class is the base class for all factories within 
 * Palamedes. It provides the management of descriptions for external classes.
 */
public abstract class AbstractFactory {

	/** Internal list of all class descriptions */
	protected List<FactoryDescription> descriptions = 
										   new ArrayList<FactoryDescription>();
	
	/**
	 * Container class for descriptions of classes to be used by any factory.
	 */
	public class FactoryDescription {

	    private final String name;
	    private final String className;
	    private final String description;

	    public FactoryDescription( String name,
	    						   String className,
	    						   String description ) {
	    	this.name 	     = name;
	    	this.className 	 = className;
	    	this.description = description;
	    }

	    public String getName() 		{ return name;	   	  }
	    public String getClassName() 	{ return className;	  }
	    public String getDescription() 	{ return description; }

	}


	/**
	 * This method adds the given description to the factory.
	 * 
	 * @param description
	 */
	public void addDescription(FactoryDescription description) {
		descriptions.add(description);
	}

	
	/**
	 * This method uses the given parameter to create a FactoryDescription 
	 * object which then will be added to the factory.
	 * 
	 * @param name
	 * @param className
	 * @param description
	 */
	public void addDescription( String name,
								String className,
								String description ) {
		addDescription( new FactoryDescription(name, className, description) );
	}

	
	/**
	 * This method removes the given description from the factory.
	 * @param description
	 */
	public void removeDescription(FactoryDescription description) {
		descriptions.remove(description);
	}
	
	
	/**
	 * This method returns the description list of the factory.
	 */
	public List<FactoryDescription> getDescriptions() {
		return descriptions;
	}


	/**
	 * This method creates an instance of a class based on the given 
	 * description.
	 * 
	 * This method is meant to be used by inherited classes and should not be
	 * public.
	 * @param description
	 * @return
	 */
	protected Object createInstance(FactoryDescription description) {
		return createInstance(description.className);
	}
	
	
	/**
	 * This method creates an instance of a class based on the given 
	 * class name.
	 * 
	 * This method is meant to be used by inherited classes and should not be
	 * public.
	 * @param className
	 * @return
	 */
	protected Object createInstance(String className) {
        try {
			return Class.forName(className).newInstance();
		} catch (InstantiationException e) {
			e.printStackTrace();
		} catch (IllegalAccessException e) {
			e.printStackTrace();
		} catch (ClassNotFoundException e) {
			e.printStackTrace();
		}
		return null;
	}

	/**
	 * This method returns the description with name \a name.
	 * @param name
	 * @return
	 */
	protected FactoryDescription getDescriptionByName(String name) {
        for (FactoryDescription description : descriptions)
            if (name.equals(description.getName()))
            	return description;
        return null;
	}
	
}
