package org.eclipse.palamedes.gdl.core.model.utils;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.eclipse.palamedes.gdl.core.model.IFluent;
import org.eclipse.palamedes.gdl.core.model.IGameNode;
import org.eclipse.palamedes.gdl.core.model.IGameState;
import org.eclipse.palamedes.gdl.core.model.IMove;


public final class GameNode implements IGameNode {

    final IGameNode  parent;
    final int        depth;
    final IMove[]    moves;
    final boolean    terminal;

    IGameState state;
    boolean    preserve;

    public GameNode() {
        parent   = null;
        moves    = null;
        depth    = 0;
        preserve = false;
        terminal = false;
    }

    public GameNode(IGameNode parent, IGameState state, IMove[] moves) {
        this.parent   = parent;
        this.state    = state;
        this.moves    = moves.clone();
        this.depth    = parent.getDepth() + 1;
        this.preserve = false;
        this.terminal = state.isTerminal();
    }

    /*
     * (non-Javadoc)
     * @see org.eclipse.palamedes.gdl.core.model.IGameNode#getParent()
     */
    public IGameNode getParent() {
        return parent;
    }

    /*
     * (non-Javadoc)
     * @see org.eclipse.palamedes.gdl.core.model.IGameNode#getDepth()
     */
    public int getDepth() {
        return depth;
    }

    /*
     * (non-Javadoc)
     * @see org.eclipse.palamedes.gdl.core.model.IGameNode#getState()
     */
    public IGameState getState() {
        return state;
    }

    /*
     * (non-Javadoc)
     * @see org.eclipse.palamedes.gdl.core.model.IGameNode#setState(org.eclipse.palamedes.gdl.core.model.IGameState)
     */
    public void setState(IGameState state) {
        this.state = state;
    }

    /*
     * (non-Javadoc)
     * @see org.eclipse.palamedes.gdl.core.model.IGameNode#getMoves()
     */
    public IMove[] getMoves() {
        return moves == null ? null : moves.clone();
    }

    /*
     * (non-Javadoc)
     * @see org.eclipse.palamedes.gdl.core.model.IGameNode#isPreserved()
     */
    public boolean isPreserved() {
        return preserve;
    }

    /*
     * (non-Javadoc)
     * @see org.eclipse.palamedes.gdl.core.model.IGameNode#setPreserve(boolean)
     */
    public void setPreserve(boolean preserve) {
        this.preserve = preserve;
    }

    /*
     * (non-Javadoc)
     * @see java.lang.Object#toString()
     */
    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append("Depth[");
        sb.append(depth);
        sb.append("] := ");
        sb.append("Moves");
        sb.append((moves == null) ? " No Move Info" : Arrays.toString(moves));
        sb.append(" => ");
        sb.append((state == null) ? "No State Info" : state.toString());
        return sb.toString();
    }

    public static int[] getDiffCount(IGameState nodeA, IGameState nodeB) {
        List<IFluent>[] lists = getDiff(nodeA, nodeB);
        return new int[] { lists[0].size(), lists[1].size(), lists[2].size() };
    }

    @SuppressWarnings("unchecked")
    public static List<IFluent>[] getDiff(IGameState nodeA, IGameState nodeB) {

        // map for comparison
        Set<IFluent>  set     = new HashSet<IFluent>();
        List<IFluent> aOnly   = new ArrayList<IFluent>();
        List<IFluent> bOnly   = new ArrayList<IFluent>();
        List<IFluent> overlap = new ArrayList<IFluent>();

        // if nodeA == null -> all nodeB fluents are positive effected
        if (nodeA != null)
            for (IFluent fluent : nodeA.getFluents())
                set.add(fluent);

        // test all fluents of nodeB against the ones of nodeA
        for ( IFluent fluent : nodeB.getFluents() )
            if (set.contains(fluent)) {
                overlap.add(fluent);
                set.remove(fluent);
            } else
                aOnly.add(fluent);

        // all fluents left in map are bOnly nodes
        bOnly.addAll(set);
        return new List[] {aOnly, bOnly, overlap};
    }

    /*
     * (non-Javadoc)
     * @see org.eclipse.palamedes.gdl.core.model.IGameNode#isTerminal()
     */
    public boolean isTerminal() {
        return terminal;
    }

}
