package org.eclipse.palamedes.gdl.core.model.utils;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Stack;

import org.eclipse.palamedes.gdl.core.model.IGameNode;
import org.eclipse.palamedes.gdl.core.model.IGameState;
import org.eclipse.palamedes.gdl.core.model.IGameTree;
import org.eclipse.palamedes.gdl.core.model.IMove;


/**
 * The GameStateTree contains the tree of already resolved game nodes.
 * @author Ingo Keller
 */
public class GameNodeTree implements IGameTree {

    /** Compacting threshold */
    private static final int cThreshold = 1000;
    
    /** Mapping ParentNode -> [ChildNode]* */
    public final HashMap<IGameNode, List<IGameNode>>     structure;

    /** Mapping: GameState.hashCode -> [GameNode.hashCode]+ */
    public final HashMap<Integer, List<Integer>>         gsHash2Nodes;

    /** Mapping: GameNode.hashCode -> GameNode */
    public final HashMap<Integer, IGameNode>             hash2Node;

    /** List: GameState.hashCode if GameState == Terminal */
    private final List<Integer>                          terminalStates;

    /** Root Node of the whole tree */
    private IGameNode                                    rootNode;

    /* Game Tree Statistic */
    int    nodeCount          = 0;
    int    jumpCount          = 0;
    double avgBranchingFactor = 0;
    int    minBranchingFactor = Integer.MAX_VALUE;
    int    maxBranchingFactor = 0;
    
    public static boolean jumpNode = false;

//    UbigraphClient        graph       = new UbigraphClient();
//    Map<Integer, Integer> edges       = new HashMap<Integer, Integer>();
//    Map<Integer, Integer> hash2Vertex = new HashMap<Integer, Integer>();
    
    /**
     * This constructor creates an empty game state tree.
     */
    public GameNodeTree() {
        structure      = new HashMap<IGameNode, List<IGameNode>>();
        gsHash2Nodes   = new HashMap<Integer,   List<Integer>>();
        hash2Node      = new HashMap<Integer,   IGameNode>();
        terminalStates = new ArrayList<Integer>();
//        graph.clear();
    }

    public void setRoot(IGameState state) {

        // add root to content mapping
        rootNode = new GameNode();
        rootNode.setState(state);
        rootNode.setPreserve(true);
        ArrayList<Integer> tmpList = new ArrayList<Integer>();
        tmpList.add(rootNode.hashCode());
        gsHash2Nodes.put( state.hashCode(), tmpList );

        // add to structure mapping
        structure.put( rootNode,            new ArrayList<IGameNode>() );
        hash2Node.put( rootNode.hashCode(), rootNode );

//        hash2Vertex.put(rootNode.hashCode(), graph.newVertex());
    }


    /**
     * This method adds a child to a parent.
     * @param parent
     * @param child
     */
    public IGameNode addChild(IGameNode parent, IGameState child, IMove[] moves) {

        // without a child we do nothing
        if (child == null)
            return null;

        // compacting tree
        if ( nodeCount++ % cThreshold == 0 )
            compactTree();
        
        // tree optimization that recognizes recurring game states
        if (jumpNode) {
            int hashCode = child.hashCode();
            if ( gsHash2Nodes.containsKey(hashCode) ) {
                jumpCount++;
                Integer   hash = gsHash2Nodes.get(hashCode).get(0);
                IGameNode node = hash2Node.get(hash);
                return getNextNode(node, moves);
            }
        }

        // create new node
        IGameNode newNode = new GameNode(parent, child, moves);

        // we check the the content mapping
        int hashCode = child.hashCode();
                

        if ( !gsHash2Nodes.containsKey(hashCode) ) {
            gsHash2Nodes.put(hashCode, new ArrayList<Integer>());

            if ( child.isTerminal() )
                terminalStates.add(hashCode);

        }
        gsHash2Nodes.get(hashCode).add(newNode.hashCode());
        hash2Node.put(newNode.hashCode(), newNode);

        // we check the presence within structure mapping
        if ( !structure.containsKey(newNode) ) {
            structure.put(newNode, new ArrayList<IGameNode>());
        
            // add vertex and edge
//            int newVertex    = graph.newVertex();
//            int parentVertex = hash2Vertex.get(parent.hashCode());
//            int edge         = graph.newEdge(parentVertex, newVertex);
            
//            hash2Vertex.put(newNode.hashCode(), newVertex);
//            edges.put(newNode.hashCode(), edge);
        }
        
        // finally we add the link between parent and child
        structure.get(parent).add(newNode);

        return newNode;
    }


    protected void getBranchingInfo() {
        avgBranchingFactor = 0;
        minBranchingFactor = Integer.MAX_VALUE;
        maxBranchingFactor = 0;
        int nbOfNodes = 0;
        double sumBranchingFactor = 0;
        for (IGameNode node : structure.keySet()) {
            if (node != null && node.getState() != null) {
                int moveCount = node.getState().getCombinedLegalMoves().size();
            	++nbOfNodes;
                sumBranchingFactor += moveCount; 
                minBranchingFactor = Math.min(minBranchingFactor, moveCount);
                maxBranchingFactor = Math.max(maxBranchingFactor, moveCount);
            }
        }
        avgBranchingFactor = sumBranchingFactor/nbOfNodes;
    }
    
    final void compactTree() {

//        System.out.println("StateCount[" +getNumberOfStates() + "] before");
        
        // compacting tree
        for (IGameNode node : structure.keySet())
            if ( !node.isPreserved() )
                node.setState(null);

//        System.out.println("StateCount[" +getNumberOfStates() + "] after");
/*
        getBranchingInfo();
        System.err.println( (nodeCount-1) + " Nodes in Tree " +
                            terminalStates.size() + "/" +
                            goalStates.size() +
                            " Terminal/Goal States in Tree " +
//                          avgBranchingFactor + " " +
//                          minBranchingFactor + " " +
                            maxBranchingFactor );
*/
    }

    protected final int getNumberOfStates() {
        int count = 0;
        for (IGameNode node : structure.keySet())
            count += (node.getState() != null) ? 1 : 0;
        return count;
    }

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder("GameStateTree:");
        sb.append(" [").append(structure.toString()).append("] ");
        return sb.toString();
    }


    public IGameNode getNextNode(IGameNode parent, IMove[] moves) {
        if ( structure.containsKey(parent) )
            for ( IGameNode node : structure.get(parent) )
                if ( Arrays.equals(node.getMoves(), moves) )
                    return node;
        return null;
    }


    public IGameNode getRootNode() {
        return rootNode;
    }

    public void setTerminal(IGameState terminalState) {
        terminalStates.add(terminalState.hashCode());
    }

    
    /**
     * Removes the subtree which has startNode as the root node.
     */
    private void removeSubTree(IGameNode startNode, boolean keepStartNode) {

        // remove start node if not kept
        if (!keepStartNode) {
            structure.remove(startNode);
            hash2Node.remove(startNode.hashCode());
        }
                
        // create stack and put starting node on it
        Stack<IGameNode> stack = new Stack<IGameNode>();
        stack.push(startNode);

        // Breadth-First-Traversal
        while ( !stack.isEmpty() ) {
            IGameNode       node       = stack.pop();
            List<IGameNode> candidates = structure.get(node);
            structure.remove(node);            
            hash2Node.remove(node.hashCode());
            
            if (candidates != null)
                for (IGameNode candidate : candidates)
                    stack.push(candidate);
        }
    }

    
    public void removeAllSiblingTrees(IGameNode node) {
//        long start = System.currentTimeMillis();
//        System.out.println(">>> Before [" + structure.size() + "/" + hash2Node.size() + "]");
        if (node != rootNode) {
        	List<IGameNode> siblings = structure.get( node.getParent() );
        	siblings.remove(node);
        	for (IGameNode sibling : siblings)
        		removeSubTree(sibling, true);
        }
//        System.out.println(">>> After [" + structure.size() +  "/" + hash2Node.size() + "]" + 
//                (System.currentTimeMillis() - start));
    }
    
}
