package org.eclipse.palamedes.gdl.core.resolver.jocular;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.eclipse.palamedes.gdl.core.model.IFluent;
import org.eclipse.palamedes.gdl.core.model.IGameState;
import org.eclipse.palamedes.gdl.core.model.IMove;
import org.eclipse.palamedes.gdl.core.model.IReasoner;
import org.eclipse.palamedes.gdl.core.model.utils.AbstractState;
import org.eclipse.palamedes.gdl.core.model.utils.GenericMove;

import stanfordlogic.game.GameManager;
import stanfordlogic.gdl.GdlList;
import stanfordlogic.gdl.Parser;
import stanfordlogic.knowledge.BasicKB;
import stanfordlogic.knowledge.GameInformation;
import stanfordlogic.knowledge.KnowledgeBase;
import stanfordlogic.knowledge.MetaGdl;
import stanfordlogic.knowledge.RelationNameProcessor;
import stanfordlogic.prover.AbstractReasoner;
import stanfordlogic.prover.BasicReasoner;
import stanfordlogic.prover.Fact;
import stanfordlogic.prover.GroundFact;
import stanfordlogic.prover.ProofContext;
import stanfordlogic.prover.TermVariable;
import stanfordlogic.prover.VariableFact;

public final class ReasonerAdapter implements IReasoner {

    private final class StateAdapter extends AbstractState<KnowledgeBase> {

        public StateAdapter(KnowledgeBase objState) {
            super(objState);
        }

        @Override
        public void initFluentList() {
            for ( GroundFact fact : nativeState.getFacts() )
                fluents.add(new FluentAdapter(reasoner.getSymbolTable(), fact.getTerm(0)));
        }

        @Override
        public String toString() {
            return nativeState.stateToGdl();
        }

    }

    private final class MoveAdapter extends GenericMove<GroundFact> {
        public MoveAdapter(GroundFact exp) {
            super(new GroundFactTermAdapter(reasoner.getSymbolTable(), exp));
        }
    }

    private final Parser                  parser;
    private final String                  source;
    private final AbstractReasoner        reasoner;

    private final Fact                    queryTerminal;
    private final Fact                    queryNext;
    private final Fact                    queryInit;
    private final RelationNameProcessor   trueProcessor;
    private final GdlList                 rules;
    private final Map<String, GroundFact> roleMap;

    public ReasonerAdapter(String gameDescription) {
        parser = GameManager.getParser();
        source = gameDescription;
        rules  = parser.parse(source);
        GameInformation gameInfo = new MetaGdl(parser).examineGdl(rules);
        KnowledgeBase   staticKb = new BasicKB();
        staticKb.loadWithFacts(gameInfo.getAllGrounds());
        reasoner = new BasicReasoner(staticKb, gameInfo.getIndexedRules(), parser);
        queryTerminal = makeQuery("terminal");
        queryNext     = makeQuery("next", "?f");
        queryInit     = makeQuery("init", "?f");
        trueProcessor = new RelationNameProcessor(parser.TOK_TRUE);

        roleMap = new HashMap<String, GroundFact>();
        for (GroundFact roleFact : reasoner.getAllAnswers(makeQuery("role", "?r")))
            roleMap.put( roleFact.getTerm(0).toString(parser.getSymbolTable()),
                         roleFact );
    }

    protected Fact makeQuery(String ... args) {
        GdlList list = GdlList.buildFromWords(parser.getSymbolTable(), args);
        return VariableFact.fromList(list);
    }

    public synchronized boolean isTerminal(ProofContext state) {
        return reasoner.getAnAnswer(queryTerminal, state) != null;
    }

    public int getGoalValue(String roleName, IGameState nextState)
    throws InterruptedException {

        Fact query = new VariableFact( parser.TOK_GOAL,
                                       roleMap.get(roleName).getTerm(0),
                                       TermVariable.makeTermVariable());

        StateAdapter state   = (StateAdapter) nextState;
        ProofContext context = new ProofContext(state.getNativeState(), parser);
        GroundFact   fact    = reasoner.getAnAnswer(query, context);

        if (fact == null) {
            System.err.println("no Fact for " + state + " -- " + context);
            return 0;
        }

        String goalString = fact.getTerm(1).toString(parser.getSymbolTable());
        return Integer.parseInt(goalString);
    }

    public IMove[] getLegalMoves(String roleName, IGameState currentNode)
    throws InterruptedException {
        Fact             question = new VariableFact(
                                               parser.TOK_LEGAL,
                                               roleMap.get(roleName).getTerm(0),
                                               TermVariable.makeTermVariable());
        StateAdapter     state    = (StateAdapter) currentNode;
        ProofContext     context  = new ProofContext( state.getNativeState(),
                                                      parser );
        List<GroundFact> facts    = reasoner.getAllAnswers(question, context);
        IMove[]          moves    = new IMove[facts.size()];
        int i = 0;
        for (GroundFact fact : facts)
            moves[i++] = new MoveAdapter(fact);
        return moves;
    }

    public IGameState getNextState(IGameState currentState, IMove[] moves)
    throws InterruptedException {

        // create new context
        StateAdapter  cuState = (StateAdapter) currentState;
        KnowledgeBase newKB   = new BasicKB();
        newKB.loadWithFacts( cuState.getNativeState().getFacts() );
        ProofContext  context = new ProofContext(newKB, parser);
        for ( IMove move : moves ) {
            GroundFact fact = ((MoveAdapter)move).getNativeMove();
            context.getVolatileKb().setTrue( new GroundFact( parser.TOK_DOES,
                                                             fact.getTerm(0),
                                                             fact.getTerm(1)) );
        }

        List<GroundFact> nextFacts = reasoner.getAllAnswers(queryNext, context);
        KnowledgeBase kb = new BasicKB();
        for (GroundFact fact : nextFacts)
            kb.setTrue(trueProcessor.processFact(fact));

        StateAdapter newState = new StateAdapter(kb);

        newKB   = new BasicKB();
        newKB.loadWithFacts( newState.getNativeState().getFacts() );
        context = new ProofContext(newKB, parser);

        if ( reasoner.getAnAnswer(queryTerminal, context) != null )
            newState.setTerminal();

        return newState;
    }

    public String[] getRoleNames() {
        List<GroundFact> roleFacts =
                               reasoner.getAllAnswers(makeQuery("role", "?r"));
        String[] roles = new String[roleFacts.size()];
        int i = 0;
        for (GroundFact roleFact : roleFacts)
            roles[i++] = roleFact.getTerm(0).toString(parser.getSymbolTable());

        return roles;
    }

    /* (non-Javadoc)
     * @see tud.gamecontroller.game.jocular.ReasonerInterface#getInitialState()
     */
     public synchronized StateAdapter getInitialState() {
         KnowledgeBase kb = new BasicKB();
         for (GroundFact init : reasoner.getAllAnswers(queryInit))
             kb.setTrue(trueProcessor.processFact(init));

         ProofContext proofContext=new ProofContext(kb, parser);
         StateAdapter state=new StateAdapter( proofContext.getVolatileKb() );
         if(isTerminal(proofContext)){
        	 state.setTerminal();
         }
         return state;
     }

    public String getSourceGDL() {
        return source;
    }

    public IFluent createFluent(String gdlString) {
        GdlList         rulesB   = parser.parse(gdlString);
        GameInformation gameInfo = new MetaGdl(parser).examineGdl(rulesB);
        return new FluentAdapter(reasoner.getSymbolTable(), gameInfo.getAllGrounds().get(0).getTerm(0) );
    }

    public IMove createMove(String gdlString) {
        GdlList         rulesB   = parser.parse(gdlString);
        GameInformation gameInfo = new MetaGdl(parser).examineGdl(rulesB);
        return new MoveAdapter( gameInfo.getAllGrounds().get(0) );
    }

}
