package org.eclipse.palamedes.gdl.core.simulation;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.palamedes.gdl.core.model.IGame;
import org.eclipse.palamedes.gdl.core.model.IGameNode;
import org.eclipse.palamedes.gdl.core.model.IMove;



public final class SimulatorIntern {

    // simulation info - static
    private final IGame           game;
    private final Match[]         matchs;

    // simulation info - dynamic
    private SimulationSettings    settings;
    private int                   step;
    private IGameNode             currentNode;
    private final IMove[]         currentMoves;
    private final List<IMove[]>   moveHistory = new ArrayList<IMove[]>();
    private final List<IGameNode> nodeHistory = new ArrayList<IGameNode>();


    public SimulatorIntern(IGame newGame) {
        super();

        // generate the matches
        game         = newGame;
        matchs       = new Match[game.getRoleCount()];
        currentMoves = new IMove[game.getRoleCount()];

        // Initialize dynamic structures
        reset();
    }

    public final void setSettings(SimulationSettings newSettings) {
        settings = newSettings;
    }

    /**
     * Creates a default random setting in the case of missing external
     * settings. Its a kind of fall-back setting.
     */
    private final void createRandomSettings() {
        int playerCount = game.getRoleCount();

        // init all arrays
        IStrategy[] strategies = new IStrategy[playerCount];
        int[]       startTimes = new int[playerCount];
        int[]       playTimes  = new int[playerCount];

        for (int i = 0; i < playerCount; i++) {
            strategies[i] = StrategyFactory.getInstance().createStrategy(0);
            startTimes[i] = -1;
            playTimes[i]  = -1;
        }

        settings = new SimulationSettings(strategies, startTimes, playTimes);
    }

    /**
     * Creates the matches and initializes the runtime information
     */
    public final void reset() {
        // create random settings if none are available
        if (settings == null)
            createRandomSettings();

        // create match objects
        for (int i = 0, j = game.getRoleCount(); i < j; i++)
            matchs[i] = new Match( game,
                                   settings.strategies[i],
                                   game.getRoleNames()[i],
                                   settings.startTimes[i],
                                   settings.playTimes[i] );

        // clear history
        nodeHistory.clear();
        moveHistory.clear();

        // set step and currentNode to start
        step         = 0;
        currentNode  = game.getTree().getRootNode();
    }

    public final void step() {

        // generate next move
        for (int i = 0, j = game.getRoleCount(); i < j; i++ )
            currentMoves[i] = matchs[i].getMove();

        // simulation bookkeeping
        nodeHistory.add(currentNode);
        moveHistory.add(currentMoves);
        step++;

        // generate next state
        for (Match match : matchs)
            match.makeTurn(currentMoves);
        currentNode = matchs[0].getCurrentNode();

        if ( isTerminal() ) {

            // empty next moves
            for (int i = 0, j = game.getRoleCount(); i < j; i++)
                currentMoves[i] = null;

            // add the last state to history
            nodeHistory.add(currentNode);
            step++;
        }
    }


    // getter for internal state
    public final List<IGameNode> getNodeHistory() {
        return nodeHistory;
    }

    public final int getStep() {
        return step;
    }

    public final boolean isTerminal() {
        return currentNode.getState().isTerminal();
    }

    public final void setCurrentNode(IGameNode currentNode) {
        this.currentNode = currentNode;
        for (Match match : matchs)
            match.currentNode = currentNode;
    }


    public List<IMove[]> getMoveHistory() {
        return moveHistory;
    }

    public Match[] getMatchs() {
        return matchs.clone();
    }
}
