package org.eclipse.palamedes.gdl.core.simulation.strategies;

import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.eclipse.palamedes.gdl.core.model.IGameNode;
import org.eclipse.palamedes.gdl.core.model.IGameState;
import org.eclipse.palamedes.gdl.core.model.IMove;
import org.eclipse.palamedes.gdl.core.simulation.Match;
import org.eclipse.palamedes.gdl.core.simulation.SimulatorIntern;
import org.eclipse.palamedes.gdl.core.simulation.TimerFlag;


public class SMonteCarlo extends AbstractStrategy {

    private static class Evaluation {
        double visitCounts = 0;
        double sumOfGoals  = 0;
    }


    SimulatorIntern          simulator;
    int                      simCount;
    int                      moveCount;
    Map<Integer, Evaluation> eval;


    @Override
    public void initMatch(Match initMatch) {
        super.initMatch(initMatch);

        // create an internal simulator
        simulator = new SimulatorIntern(game);
        eval      = new HashMap<Integer, Evaluation>();
        simCount  = 0;
        moveCount = 0;

        simulate();
    }

    private void simulate() {

        while (true) {

            // reset simulation -> random settings are used
            simulator.reset();
            simulator.setCurrentNode(match.getCurrentNode());

            TimerFlag timer = match.getTimer();
            while ( !simulator.isTerminal() )
                if ( timer.interrupted() )
                    return;
                else
                    simulator.step();

            saveEvaluation();
            simCount++;
        }
    }


    private void saveEvaluation() {

        // get my goal
        List<IGameNode> nodes     = simulator.getNodeHistory();
        IGameState      lastState = nodes.get( nodes.size() - 1 ).getState();
        int             myGoal    = lastState.getGoalValue(playerNumber);

        // back propagate the goal
        for (int i = nodes.size() -1; i >= 0; i--) {
            IGameNode node     = nodes.get(i);
            int       hashCode = node.hashCode();

            // make sure we have a evaluation object for the node
            if ( !eval.containsKey(hashCode) )
                eval.put(hashCode, new Evaluation());

            Evaluation evaluation = eval.get(hashCode);
            evaluation.visitCounts++;
            evaluation.sumOfGoals += myGoal;

            // we don't want to keep all the simulated states
            if (moveCount < i)
                node.setPreserve(false);
        }
    }


    @Override
    public IMove getMove(IGameNode currentNode) {
        simulate();

        try {
            // find best child node
            List<IMove[]> moveList    = game.getCombinedMoves(currentNode);
            IMove[]       bestMove    = null;
            IGameNode     bestNode    = null;
            double        bestAvgGoal = 0;

            // iterate over all moves
            for (IMove[] move : moveList) {
                IGameNode nextNode = game.getNextNode(currentNode, move);

                // we only look for already resolved nodes
                if (nextNode == null)
                    continue;

                // no best node so far
                if (bestNode == null) {
                    bestNode    = nextNode;
                    bestMove    = move;

                    Evaluation evalNext = eval.get(nextNode.hashCode());
                    if (evalNext == null)
                        bestAvgGoal = 0;
                    else
                        bestAvgGoal = evalNext.sumOfGoals / evalNext.visitCounts;
                    continue;
                }

                // skip if node was not already evaluated
                Evaluation evalNext = eval.get(nextNode.hashCode());
                if (evalNext == null)
                    continue;

                // did we found something better ?
                double nextAvgGoal = evalNext.sumOfGoals / evalNext.visitCounts;
                if ( nextAvgGoal > bestAvgGoal) {
                    bestAvgGoal = nextAvgGoal;
                    bestMove    = move;
                    bestNode    = nextNode;
                }

            }

            if (bestMove == null) {
                System.out.println("Error --- Available Moves");
                for (IMove[] move : game.getCombinedMoves(currentNode))
                    System.out.println(Arrays.toString(move));
                return null;
            }

            moveCount++;
//            System.out.println( ">>> SMonteCarlo Move Selection");
//            System.out.println( "bestGoalValue["      + bestAvgGoal +
//                                "], move["            + bestMove[playerNumber] +
//                                "], simulationCount[" + simCount +"]");
            return bestMove[playerNumber];
        }
        catch (InterruptedException e) {
        }
        return null;
    }

}
